/** 
 *   Encoders: Control over iar bus (server).
 *
 *   (c) Copyright 205 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: 
 *        exit on encoders_core();
 **/
#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <syslog.h>
#include <signal.h>

#include "e5850.h"
#include "my_stdio.h"
#include "iar_bus.h"
#include "antenna_ib_pk.h"
#define ENCODERS_HW
#include "encoders.h"


int e5850_nrb = -1;
int encoder_num[] = {RA_CASE, DEC_CASE};
//char antenna_hostname[_POSIX_PATH_MAX] = {ANTENNA_HOST_NAME};

struct encoders_data *d = NULL;
int encoders_done = 0;

/*****************************************************************************
 * local functions:
 **/

static ssize_t soft_read( int fd, void *buf, size_t count)
{
        fd_set rfds;
        int fd_max = fd + 1;
        int retval;
	struct timeval tv;

	while( 1 ) {
		tv.tv_sec = 10;
		tv.tv_usec = 0;
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);
		retval = select(fd_max, &rfds, NULL, NULL, &tv);
		if (retval) {
			if( FD_ISSET(fd, &rfds) ) {
				retval = read(fd, buf, count);
#ifdef DEBUG_MODE
				if (verbose_flag & PINFO) {
					int __i;
					PVERB(PINFO, 
					      "# encoders_ib_server.c read <- ");
					for(__i=0;__i<retval;__i++)
						PVERB(PINFO, "%02x ",
						      *(((u_char *)buf)+__i) );
					PVERB(PINFO, "\n");
				}
#endif
				return retval;
			}
		}
	}		
	
	return -1; /* never */
}
	

static inline int gray2bin(int bits, int binary)
{
	int gray_code = 0;
	int bit_number;
	int bit = 0;
	
	for (bit_number = bits-1; bit_number >= 0; bit_number--) {
		bit ^= binary >> bit_number & 1;
		if(bit)
			gray_code |= (1 << bit_number);
	}

	return gray_code;
}
	

static int get_bits_off_resolution(int fd)
{
	if(fd)
		return ioctl(fd, E5850_GETBITRES);
	return -1;
}


static inline int diff_evaluate(int old, int new, int *current_factor)
{
	int diff = old - new;
	
/* FIXME: el ruido es 0x1fff, 0xfff, 0x7ff, 0x3ff */	

	if(diff < - 0x2000) { /* 2^13 transit from W to E or from S to N */
		*current_factor -= 0x4000;
		/*
		printf("factor(-): %d, 0x%x [%d - %d]\n", 
		       (unsigned int) *current_factor, 
		       (unsigned int) *current_factor,
		       old, new);
		*/
	}
	if (diff > 0x2000) {
		*current_factor += 0x4000;
		/*
		printf("factor(+): %d, 0x%x [%d - %d]\n", 
		       (unsigned int) *current_factor, 
		       (unsigned int) *current_factor,
		       old, new);
		*/
	}

	return *current_factor + new;
}


static int encoders_get_next_data(int fd, struct encoders_data *en)
{
	int i;
	char buff[E5850_DATA_LEN];
	char *np = buff;
	char *endp;

#ifndef DEMO_MODE
	if ( read(fd, buff, E5850_DATA_LEN) != E5850_DATA_LEN)
		return -1;
#else
	usleep(240000);
	sprintf(buff,"0001 1ffe 0050\n");
#endif
	for(i=0;i<2;i++){
		d->encoder[i] = 
			gray2bin(e5850_nrb, 
				 (int) strtol(np, &endp, 16 ));
		d->val[i] = diff_evaluate(d->old[i], d->encoder[i],
					  &d->factor[i]);
		d->old[i] = d->encoder[i];

		np = endp + 1;
	}
#if USE_ANEMOMETER
	d->anemometer = (int) strtol(np, &endp, 16 );
#endif
	if( en != NULL) memcpy(en, d, sizeof(struct encoders_data));

	return 1;
}


static int encoders_cmd(int fd, int fd_bus)
{
	struct ib_packet pk;
	struct antenna_packet *anp;
	int ret;
	int val = -1;
	int len;
	int send_ret;

	anp = (struct antenna_packet *)&pk.data;
	send_ret = 1;
	memset(&pk, 0, sizeof(struct ib_packet));
	ret = soft_read(fd_bus, &pk, sizeof(struct ib_packet_head) );	
	if( pk.h.size > 0) {
		ret += soft_read(fd_bus, &pk.data, pk.h.size );
		val = atoi(anp->data);
	}
	switch (anp->h.cmd) {
	case ECM_GETBUFFLEN:
		ioctl(fd, E5850_GETBUFFLEN, &ret);
		ret = ret >> 4;
		break;
	case ECM_CLEARBUFF:
		ret = ioctl(fd, E5850_CLEARBUFF);
		break;
	case ECM_START:
		ret = ioctl(fd, E5850_START);
		break;
	case ECM_STOP:
		ret = ioctl(fd, E5850_STOP);
		break;
	case ECM_ISRUNNING:
		ret = ioctl(fd, E5850_ISRUNNING);
		break;
	case ECM_GETNVAL:
		ret = ioctl(fd, E5850_GETNVAL, &val);
		if( ret == 0 )
			ret = gray2bin(e5850_nrb, val);
		break;
	case ECM_SETCW:
		ret = ioctl(fd, E5850_SETCW, &val);
		break;
	case ECM_SETCCW:
		ret = ioctl(fd, E5850_SETCCW, &val);
		break;
	case ECM_SETZERO:
		ret = ioctl(fd, E5850_SETZERO, &val);
		break;
	case ECM_TOGGLEZERO:
		ret = ioctl(fd, E5850_TOGGLEZERO, &val);
		break;
	case ECM_SCLKRUNNING:
		ret = ioctl(fd, E5850_SCLKRUNNING);
		break;
	case ECM_FAILMODE:
		ret = ioctl(fd, E5850_FAILMODE);
		break;
	case ECM_NORMALMODE:
		ret = ioctl(fd, E5850_NORMALMODE);
		break;
	default:
		SYSLOG(LOG_ERR,
		       " encoders core: bad command (0x%x) from 0x%x",
		       anp->h.cmd, pk.h.addr);
	case WCM_WIND: /* ignored */
	case SCM_INFO: /* ignored */
		send_ret = 0;
		break;
	}
	if ( send_ret) {
		/* Send ret value: */
		PVERB(PINFO,"# sendind ret: %d\n", ret);
		len = sizeof(struct antenna_packet_head);
		anp->h.size = sprintf(anp->data, "%d",  ret);
		len += anp->h.size;
		anp->h.cmd = CM_RETURN;
		ibus_send_data_to(fd_bus, pk.h.addr, anp, len);
	}

	return 0;
}

/*
 * local functions END.
 *****************************************************************************/


int encoders_open(int ra, int dec)
{
	int i;
	int fd;

	if ((fd = open("/dev/e5850", O_RDONLY)) < 0) {
		SYSLOG(LOG_ERR, "Error opening /dev/e5850");
		return -1;
	}
	encoder_num[RA_CASE] = ra;
	encoder_num[DEC_CASE] = dec;
	e5850_nrb = get_bits_off_resolution(fd);
	d = (struct encoders_data *) 
		malloc(sizeof(struct encoders_data));
	for(i=0;i<2;i++){
		d->factor[i] = 0;
		d->old[i] = 0;
	}

	return fd;
}


int encoders_close(int fd)
{
	return close(fd);
}



int encoders_core(int fd, int fd_bus)
{
	struct antenna_packet ap;
	int data_len = sizeof(struct antenna_packet_head) +  
		sizeof(struct encoders_data);
	fd_set rfds;
        int fd_max = fd_bus + 1;
        int retval;
	struct timeval tv;

	ap.h.cmd = ECM_POSITION;
	while(!encoders_done){
		encoders_get_next_data(fd, (struct encoders_data *)&ap.data );
		ibus_send_data( fd_bus, &ap, data_len );
		tv.tv_sec = 0;
		tv.tv_usec = 1000;
		FD_ZERO(&rfds);
		FD_SET(fd_bus, &rfds);
		retval = select(fd_max, &rfds, NULL, NULL, &tv);
		if (retval) {
			encoders_cmd(fd, fd_bus);
		}
	}

	return 0;
}


